package com.yourkit.probes.builtin;

import com.yourkit.probes.*;
import com.yourkit.runtime.Callback;

import java.io.*;
import java.lang.reflect.Field;
import java.util.Map;

@SuppressWarnings({"UnusedDeclaration"})
public final class Processes {

  /**
   * Mapping between tracked object (process or stream) to row index in T_PROCESS table
   */
  private static final ObjectRowIndexMap<Object> ourObject2ProcessRowIndex = new ObjectRowIndexMap<Object>();
  /**
   * Remember stderr streams to distinct them from stdout streams
   */
  private static final ObjectSet<Object> ourErrorStreams = new ObjectSet<Object>();

  private static final class ProcessTable extends Table {
    private final StringColumn myCommand = new StringColumn("Command");
    private final StringColumn myDirectory = new StringColumn("Directory");
    private final StringColumn myEnvironment = new StringColumn("Environment");
    private final IntColumn myExitCode = new IntColumn("Exit Code");

    public ProcessTable() {
      super("Processes", LASTING_EVENTS | RECORD_THREAD_ON_ROW_CREATION | RECORD_STACKTRACE_ON_ROW_CREATION );
    }
  }
  private static final ProcessTable T_PROCESS = new ProcessTable();

  private static final class IOOperationTable extends Table {
    private final ForeignKeyColumn myProcess = new ForeignKeyColumn(T_PROCESS);
    private final IntColumn myBytes = new IntColumn("Bytes");

    public IOOperationTable(final String title) {
      super(title, Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final IOOperationTable T_STDOUT = new IOOperationTable("Output Reads");
  private static final IOOperationTable T_STDERR = new IOOperationTable("Error Reads");
  private static final IOOperationTable T_STDIN = new IOOperationTable("Input Writes");

  static {
    // A daemon thread to asynchronously monitor exited processes.
    // We need it to handle processes for which waitFor() is not used.
    final Thread thread = new Thread(
      new Runnable() {
        public void run() {
          //noinspection InfiniteLoopStatement
          while (true) {
            try {
              try {
                Thread.sleep(1000);
              }
              catch (InterruptedException ignored) {
                continue;
              }

              final int[] exitCodes = JVM.getExitedProcesses();
              if (exitCodes == null) {
                continue;
              }

              for (int i = 0; i < exitCodes.length; i += 2) {
                final int rowIndex = exitCodes[i];
                final int exitCode = exitCodes[i+1];

                handleProcessExit(rowIndex, exitCode);
              }
            }
            catch (final Throwable e) {
              Callback.messageToLogFile(Callback.getExceptionStackTrace(e));
            }
          }
        }
      },
      "YJPAgent-Process-Exit-Listener"
    );
    thread.setDaemon(true);
    thread.start();
  }

  private static void handleProcessExit(final int rowIndex, final int exitCode) {
    if (rowIndex == Table.NO_ROW) {
      return;
    }

    if (T_PROCESS.isRowClosed(rowIndex)) {
      // already closed
      return;
    }

    T_PROCESS.myExitCode.setValue(rowIndex, exitCode);
    T_PROCESS.closeRow(rowIndex);
  }

  /**
   * Associate input streams with process
   */
  @MethodPattern(
    {
      "java.lang.ProcessImpl:getInputStream()",
      "java.lang.UNIXProcess:getInputStream()"
    }
  )
  public static final class ProcessImpl_getInputStream_Probe {
    public static void onReturn(
      @ReturnValue final InputStream stream,
      @This final Process process
    ) {
      final int rowIndex = ourObject2ProcessRowIndex.get(process);
      if (rowIndex == Table.NO_ROW) {
        return;
      }

      final InputStream inputStream = getProperInputStream(stream);
      if (inputStream == null) {
        return;
      }

      ourObject2ProcessRowIndex.put(inputStream, rowIndex);
    }
  }

  /**
   * Associate error streams with process
   */
  @MethodPattern(
    {
      "java.lang.ProcessImpl:getErrorStream()",
      "java.lang.UNIXProcess:getErrorStream()"
    }
  )
  public static final class ProcessImpl_getErrorStream_Probe {
    public static void onReturn(
      @ReturnValue final InputStream stream,
      @This final Process process
    ) {
      final int rowIndex = ourObject2ProcessRowIndex.get(process);
      if (rowIndex == Table.NO_ROW) {
        return;
      }

      final InputStream errorStream = getProperInputStream(stream);
      if (errorStream == null) {
        return;
      }

      ourObject2ProcessRowIndex.put(errorStream, rowIndex);
      ourErrorStreams.add(errorStream);
    }
  }

  /**
   * Associate output streams with process
   */
  @MethodPattern(
    {
      "java.lang.ProcessImpl:getOutputStream()",
      "java.lang.UNIXProcess:getOutputStream()"
    }
  )
  public static final class ProcessImpl_getOutputStream_Probe {
    public static void onReturn(
      @ReturnValue final OutputStream stream,
      @This final Process process
    ) {
      final int rowIndex = ourObject2ProcessRowIndex.get(process);
      if (rowIndex == Table.NO_ROW) {
        return;
      }

      final FileOutputStream outputStream = getFileOutputStream(stream);
      if (outputStream == null) {
        return;
      }

      ourObject2ProcessRowIndex.put(outputStream, rowIndex);
    }
  }

  @MethodPattern("java.lang.UNIXProcess:waitForProcessExit(int)")
  public static final class UnixProcess_waitForProcessExit_Probe {
    public static void onReturn(
      @ReturnValue final int exitCode,
      @This final Process process
    ) {
      final int rowIndex = ourObject2ProcessRowIndex.get(process);
      handleProcessExit(rowIndex, exitCode);
    }
  }

  @MethodPattern("java.lang.ProcessImpl:start(String[], java.util.Map, String, *)")
  public static final class ProcessImpl_start_Probe {
    public static void onReturn(
      @ReturnValue final Process process,
      @Param(1) final String[] cmd,
      @Param(2) final Map<String,String> env,
      @Param(3) final String dir
    ) {
      final int rowIndex = T_PROCESS.createRow();
      T_PROCESS.myCommand.setValue(rowIndex, getCommandLine(cmd));
      T_PROCESS.myDirectory.setValue(rowIndex, dir);
      T_PROCESS.myEnvironment.setValue(rowIndex, getEnvironmentAsString(env));

      ourObject2ProcessRowIndex.put(process, rowIndex);

      // ensure that we record the process exit code
      JVM.registerProcessExitListener(process, rowIndex);
    }
  }

  @MethodPattern(
    {
      "java.lang.ProcessImpl:waitFor()",
      "java.lang.UNIXProcess:waitFor()"
    }
  )
  public static final class ProcessImpl_waitFor_Probe {
    public static void onReturn(
      @This final Process process,
      @ReturnValue final int exitValue
    ) {
      final int rowIndex = ourObject2ProcessRowIndex.get(process);
      handleProcessExit(rowIndex, exitValue);
    }
  }

  @MethodPattern("java.lang.ProcessImpl:finalize()")
  public static final class ProcessImpl_close_Probe {
    public static void onEnter(@This final Process process) {
      final int exitValue;
      try {
        exitValue = process.exitValue();
      }
      catch (Throwable ignored){
        // the process did not exit yet
        return;
      }

      final int rowIndex = ourObject2ProcessRowIndex.remove(process);
      handleProcessExit(rowIndex, exitValue);
    }
  }

  @MethodPattern("java.io.FileOutputStream:writeBytes(byte[], int, int)")
  public static final class FileOutputStream_writeBytes_Probe {
    public static int onEnter(
      @This final FileOutputStream fileOutputStream,
      @Param(3) final int bytesWritten
    ) {
      return processWriteEnterImpl(fileOutputStream, bytesWritten);
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      processWriteReturnImpl(rowIndex);
    }
  }

  @MethodPattern("java.io.FileOutputStream:write(int)")
  public static final class FileOutputStream_write_Probe {
    public static int onEnter(@This final FileOutputStream fileOutputStream) {
      return processWriteEnterImpl(fileOutputStream, 1 /* one byte to be written */);
    }

    public static void onReturn(@OnEnterResult final int rowIndex) {
      processWriteReturnImpl(rowIndex);
    }
  }

  private static int processWriteEnterImpl(final FileOutputStream fileOutputStream, final int bytesWritten) {
    final int processRowIndex = ourObject2ProcessRowIndex.get(fileOutputStream);
    if (processRowIndex == Table.NO_ROW) {
      return -1;
    }

    final int writeRowIndex = T_STDIN.createRow();
    T_STDIN.myBytes.setValue(writeRowIndex, bytesWritten);
    T_STDIN.myProcess.setValue(writeRowIndex, processRowIndex);
    return writeRowIndex;
  }

  private static void processWriteReturnImpl(final int rowIndex) {
    if (rowIndex != -1) {
      T_STDIN.closeRow(rowIndex);
    }
  }

  @MethodPattern(
    {
      "java.io.FileInputStream:read()",
      "java.lang.ProcessInputStream:read()"
    }
  )
  public static final class InputStream_read_Probe {
    public static int onEnter(@This final InputStream inputStream) {
      return processReadEnterImpl(inputStream);
    }

    public static void onReturn(
      @This final InputStream inputStream,
      @OnEnterResult final int readRowIndex,
      @ReturnValue final int returnValue
    ) {
      // non negative return value means that a byte has been successfully read
      final int readBytes = returnValue >= 0 ? 1 : 0;
      processReadReturnImpl(inputStream, readRowIndex, readBytes);
    }
  }

  @MethodPattern(
    {
      "java.io.FileInputStream:readBytes(byte[], int, int)",
      // The following is for IBM Java on UNIX:
      "java.lang.ProcessInputStream:read(byte[], int, int)",
      "java.lang.ProcessInputStream:read(byte[])"
    }
  )
  public static final class InputStream_readBytes_Probe {
    public static int onEnter(@This final InputStream inputStream) {
      return processReadEnterImpl(inputStream);
    }

    public static void onReturn(
      @This final InputStream inputStream,
      @OnEnterResult final int readRowIndex,
      @ReturnValue final int bytesRead
    ) {
      processReadReturnImpl(inputStream, readRowIndex, bytesRead);
    }
  }

  private static int processReadEnterImpl(final InputStream inputStream) {
    final int processRowIndex = ourObject2ProcessRowIndex.get(inputStream);
    if (processRowIndex == Table.NO_ROW) {
      return -1;
    }

    final IOOperationTable readTable = ourErrorStreams.contains(inputStream) ? T_STDERR : T_STDOUT;

    // log individual read event
    final int readRowIndex = readTable.createRow();
    readTable.myProcess.setValue(readRowIndex, processRowIndex);
    return readRowIndex;
  }

  private static void processReadReturnImpl(
    final InputStream inputStream,
    final int readRowIndex,
    final int bytesRead
  ) {
    if (readRowIndex == -1) {
      return;
    }

    final IOOperationTable readTable = ourErrorStreams.contains(inputStream) ? T_STDERR : T_STDOUT;

    if (bytesRead >= 0) {
      readTable.myBytes.setValue(readRowIndex, bytesRead);
    }
    readTable.closeRow(readRowIndex);
  }

  private static FileOutputStream getFileOutputStream(final OutputStream stream) {
    if (stream instanceof BufferedOutputStream) {
      try {
        final Field field = stream.getClass().getSuperclass().getDeclaredField("out");
        field.setAccessible(true);
        return (FileOutputStream)field.get(stream);
      }
      catch (Throwable ignored) {
        return null;
      }
    }

    if (stream instanceof FileOutputStream) {
      return (FileOutputStream)stream;
    }

    return null;
  }

  private static InputStream getProperInputStream(final InputStream stream) {
    if (stream instanceof BufferedInputStream) {
      try {
        final Field field = stream.getClass().getSuperclass().getDeclaredField("in");
        field.setAccessible(true);
        return (InputStream)field.get(stream);
      }
      catch (Throwable ignored) {
        return null;
      }
    }

    if (stream instanceof FileInputStream) {
      return stream;
    }

    if (
      stream != null &&
      // cannot use instanceof for this non-public IBM Java class
      "java.lang.ProcessInputStream".equals(stream.getClass().getName())
    ) {
      return stream;
    }

    return null;
  }

  private static String getCommandLine(final String[] cmd) {
    final StringBuilder cmdbuf = new StringBuilder(80);
    for (int i = 0; i < cmd.length; i++) {
      if (i > 0) {
        cmdbuf.append(' ');
      }
      final String s = cmd[i];
      if (s.indexOf(' ') >= 0 || s.indexOf('\t') >= 0) {
        if (s.charAt(0) != '"') {
          cmdbuf.append('"');
          cmdbuf.append(s);
          if (s.endsWith("\\")) {
            cmdbuf.append("\\");
          }
          cmdbuf.append('"');
        }
        else if (s.endsWith("\"")) {
          // The argument has already been quoted
          cmdbuf.append(s);
        }
      }
      else {
        cmdbuf.append(s);
      }
    }
    return cmdbuf.toString();
  }

  static String getEnvironmentAsString(final java.util.Map<String,String> env) {
    if (env == null) {
      return "";
    }

    final StringBuilder sb = new StringBuilder();
    for (final Map.Entry<String,String> e : env.entrySet()) {
      sb.append(e.getKey());
      sb.append('=');
      sb.append(e.getValue());
      sb.append('\n');
    }

    return sb.toString();
  }
}
