package com.yourkit.probes.builtin;

import com.yourkit.probes.*;

import java.io.RandomAccessFile;

public final class RandomAccessFiles {
  /**
   * Mapping between random access file and row index in T_RANDOM_ACCESS_FILE table
   */
  private static final ObjectRowIndexMap<RandomAccessFile> ourRandomAccessFile2RowIndex = new ObjectRowIndexMap<RandomAccessFile>();

  private static final class FileTable extends Table {
    private final StringColumn myFileName = new StringColumn("File Name");

    public FileTable() {
      super("Random Access Files", Table.MASK_FOR_GENERIC_LASTING_EVENTS);
    }
  }
  private static final FileTable T_RANDOM_ACCESS_FILE = new FileTable();

  private static final class FailedFileTable extends Table {
    private final StringColumn myFileName = new StringColumn("File Name");

    public FailedFileTable() {
      super("Failed Random Access Files", Table.MASK_FOR_POINT_EVENTS);
    }
  }
  private static final FailedFileTable T_FAILED_RANDOM_ACCESS_FILE = new FailedFileTable();

  private static final class ReadTable extends Table {
    private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_RANDOM_ACCESS_FILE);
    private final IntColumn myBytes = new IntColumn("Bytes");

    public ReadTable() {
      super("Reads", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final ReadTable T_READ = new ReadTable();

  private static final class WriteTable extends Table {
    private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_RANDOM_ACCESS_FILE);
    private final IntColumn myBytes = new IntColumn("Bytes");

    public WriteTable() {
      super("Writes", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
    }
  }
  private static final WriteTable T_WRITE = new WriteTable();

  @MethodPattern("java.io.RandomAccessFile:open(String, int)")
  public static final class RandomAccessFile_open_Probe {
    public static void onReturn(
      @This final RandomAccessFile randomAccessFile,
      @Param(1) final String fileName
    ) {
      final int rowIndex = T_RANDOM_ACCESS_FILE.createRow();
      T_RANDOM_ACCESS_FILE.myFileName.setValue(rowIndex, fileName);
      ourRandomAccessFile2RowIndex.put(randomAccessFile, rowIndex);
    }

    public static void onUncaughtException(
      @Param(1) final String fileName
    ) {
      final int rowIndex = T_FAILED_RANDOM_ACCESS_FILE.createRow();
      T_FAILED_RANDOM_ACCESS_FILE.myFileName.setValue(rowIndex, fileName);
    }
  }

  @MethodPattern("java.io.RandomAccessFile:close()")
  public static final class RandomAccessFile_close_Probe {
    public static void onReturn(@This final RandomAccessFile randomAccessFile) {
      final int rowIndex = ourRandomAccessFile2RowIndex.remove(randomAccessFile);
      if (rowIndex == Table.NO_ROW) {
        return;
      }

      T_RANDOM_ACCESS_FILE.closeRow(rowIndex);
    }
  }

  @MethodPattern("java.io.RandomAccessFile:writeBytes(byte[], int, int)")
  public static final class RandomAccessFile_writeBytes_Probe {
    public static int onEnter(
      @This final RandomAccessFile randomAccessFile,
      @Param(3) final int bytesWritten
    ) {
      return processWriteEnterImpl(randomAccessFile, bytesWritten);
    }

    public static void onReturn(@OnEnterResult final int writeRowIndex) {
      processWriteReturnImpl(writeRowIndex);
    }
  }

  @MethodPattern("java.io.RandomAccessFile:write(int)")
  public static final class RandomAccessFile_write_Probe {
    public static int onEnter(@This final RandomAccessFile randomAccessFile) {
      return processWriteEnterImpl(randomAccessFile, 1 /* one byte to write */);
    }

    public static void onReturn(@OnEnterResult final int writeRowIndex) {
      processWriteReturnImpl(writeRowIndex);
    }
  }

  private static int processWriteEnterImpl(final RandomAccessFile randomAccessFile, final int bytesWritten) {
    final int fileRowIndex = ourRandomAccessFile2RowIndex.get(randomAccessFile);
    if (fileRowIndex == Table.NO_ROW) {
      return -1;
    }

    // log individual write event
    final int writeRowIndex = T_WRITE.createRow();
    T_WRITE.myBytes.setValue(writeRowIndex, bytesWritten);
    T_WRITE.myFile.setValue(writeRowIndex, fileRowIndex);
    return writeRowIndex;
  }

  private static void processWriteReturnImpl(final int writeRowIndex) {
    if (writeRowIndex != -1) {
      T_WRITE.closeRow(writeRowIndex);
    }
  }

  @MethodPattern("java.io.RandomAccessFile:read()")
  public static final class RandomAccessFile_read_Probe {
    public static int onEnter(@This final RandomAccessFile randomAccessFile) {
      return processReadEnterImpl(randomAccessFile);
    }

    public static void onReturn(
      @OnEnterResult final int readRowIndex,
      @ReturnValue final int returnValue
    ) {
      // non negative return value means that a byte has been successfully read
      final int readBytes = returnValue >= 0 ? 1 : 0;
      processReadReturnImpl(readRowIndex, readBytes);
    }
  }

  @MethodPattern("java.io.RandomAccessFile:readBytes(byte[], int, int)")
  public static final class RandomAccessFile_readBytes_Probe {
    public static int onEnter(@This final RandomAccessFile randomAccessFile) {
      return processReadEnterImpl(randomAccessFile);
    }

    public static void onReturn(
      @OnEnterResult final int readRowIndex,
      @ReturnValue final int bytesRead
    ) {
      processReadReturnImpl(readRowIndex, bytesRead);
    }
  }

  private static int processReadEnterImpl(final RandomAccessFile randomAccessFile) {
    final int fileRowIndex = ourRandomAccessFile2RowIndex.get(randomAccessFile);
    if (fileRowIndex == Table.NO_ROW) {
      return -1;
    }

    // log individual read event
    final int readRowIndex = T_READ.createRow();
    T_READ.myFile.setValue(readRowIndex, fileRowIndex);
    return readRowIndex;
  }

  private static void processReadReturnImpl(final int readRowIndex, final int bytesRead) {
    if (readRowIndex == -1) {
      return;
    }

    if (bytesRead >= 0) {
      T_READ.myBytes.setValue(readRowIndex, bytesRead);
    }
    T_READ.closeRow(readRowIndex);
  }

  public static FileChannels.ChannelHandler registerChannel(final RandomAccessFile randomAccessFile) {
    final int rowIndex = ourRandomAccessFile2RowIndex.get(randomAccessFile);
    if (rowIndex == Table.NO_ROW) {
      return null;
    }
    return new MyChannelHandler(rowIndex);
  }

  static class MyChannelHandler implements FileChannels.ChannelHandler {
    final int myFileRowIndex;

    private static final class ChannelReadTable extends Table {
      private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_RANDOM_ACCESS_FILE);
      private final LongColumn myBytes = new LongColumn("Bytes");

      public ChannelReadTable() {
        super("Channel Reads", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
      }
    }
    private static final ChannelReadTable T_CHANNEL_READ = new ChannelReadTable();

    private static final class ChannelWriteTable extends Table {
      private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_RANDOM_ACCESS_FILE);
      private final LongColumn myBytes = new LongColumn("Bytes");

      public ChannelWriteTable() {
        super("Channel Writes", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
      }
    }
    private static final ChannelWriteTable T_CHANNEL_WRITE = new ChannelWriteTable();

    private static final class TransferFromTable extends Table {
      private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_RANDOM_ACCESS_FILE);
      private final LongColumn myBytes = new LongColumn("Bytes");

      public TransferFromTable() {
        super("Transfers from Channel", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
      }
    }
    private static final TransferFromTable T_TRANSFER_FROM_CHANNEL = new TransferFromTable();

    private static final class TransferToTable extends Table {
      private final ForeignKeyColumn myFile = new ForeignKeyColumn(T_RANDOM_ACCESS_FILE);
      private final LongColumn myBytes = new LongColumn("Bytes");

      public TransferToTable() {
        super("Transfers to Channel", Table.MASK_FOR_SINGLE_METHOD_CALL_LASTING_EVENTS);
      }
    }
    private static final TransferToTable T_TRANSFER_TO_CHANNEL = new TransferToTable();

    private MyChannelHandler(final int parentRowIndex) {
      myFileRowIndex = parentRowIndex;
    }

    public int readOnEnter() {
      final int readRowIndex = T_CHANNEL_READ.createRow();
      T_CHANNEL_READ.myFile.setValue(readRowIndex, myFileRowIndex);
      return readRowIndex;
    }

    public void readOnReturn(final int readRowIndex, final long bytesRead) {
      T_CHANNEL_READ.myBytes.setValue(readRowIndex, bytesRead);
      T_CHANNEL_READ.closeRow(readRowIndex);
    }

    public int writeOnEnter() {
      final int writeRowIndex = T_CHANNEL_WRITE.createRow();
      T_CHANNEL_WRITE.myFile.setValue(writeRowIndex, myFileRowIndex);
      return writeRowIndex;
    }

    public void writeOnReturn(final int writeRowIndex, final long bytesWritten) {
      if (bytesWritten >= 0) {
        T_CHANNEL_WRITE.myBytes.setValue(writeRowIndex, bytesWritten);
      }
      T_CHANNEL_WRITE.closeRow(writeRowIndex);
    }

    public int transferToOnEnter() {
      final int transferToRowIndex = T_TRANSFER_TO_CHANNEL.createRow();
      T_TRANSFER_TO_CHANNEL.myFile.setValue(transferToRowIndex, myFileRowIndex);
      return transferToRowIndex;
    }

    public void transferToOnReturn(final int transferToRowIndex, final long transferredBytes) {
      T_TRANSFER_TO_CHANNEL.myBytes.setValue(transferToRowIndex, transferredBytes);
      T_TRANSFER_TO_CHANNEL.closeRow(transferToRowIndex);
    }

    public int transferFromOnEnter() {
      final int transferFromRowIndex = T_TRANSFER_FROM_CHANNEL.createRow();
      T_TRANSFER_FROM_CHANNEL.myFile.setValue(transferFromRowIndex, myFileRowIndex);
      return transferFromRowIndex;
    }

    public void transferFromOnReturn(final int transferFromRowIndex, final long transferredBytes) {
      T_TRANSFER_FROM_CHANNEL.myBytes.setValue(transferFromRowIndex, transferredBytes);
      T_TRANSFER_FROM_CHANNEL.closeRow(transferFromRowIndex);
    }
  }
}
